#!/usr/bin/env python3

# TODO docs

import logging
from argparse import ArgumentParser
from logging import StreamHandler
from logging import getLogger
from sys import stdout

from dfmpy.commands.initialize import initialize
from dfmpy.commands.ls import ls
from dfmpy.commands.remove import remove
from dfmpy.commands.sync import sync
from dfmpy.utils.interactive import check_for_interactivity
from dfmpy.utils.logging import get_formatter

# TODO docs
LOG = getLogger()

# TODO docs
__LOG_LEVELS = tuple([
    logging.CRITICAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
])


def __setup_logger(verbosity_index=None):
    # TODO docs
    # TODO unit test
    if not verbosity_index:
        verbosity_index = 1
    verbosity_index = max(0, min(len(__LOG_LEVELS) - 1, verbosity_index))
    log_level = __LOG_LEVELS[verbosity_index]
    # formatter = Formatter('%(asctime)s %(levelname)s %(message)s')
    formatter = get_formatter('%(levelname)s %(message)s')
    handler = StreamHandler(stdout)
    handler.setFormatter(formatter)
    logger = getLogger()
    logger.setLevel(log_level)
    logger.addHandler(handler)


def __parse_arguments():
    # TODO docs
    # TODO unit test

    verbose_args = tuple(['-v', '--verbose'])
    verbose_kwargs = {
        'action': 'count',
        'default': 1,
        'help': 'Set verbosity level, multiple flags increase the level.',
    }
    force_args = tuple(['-f', '--force'])
    force_kwargs = {
        'action': 'store_true',
        'default': False,
        'dest': 'force',
    }
    interactive_args = tuple(['-i', '--interactive'])
    interactive_kwargs = {
        'action': 'store_true',
        'default': False,
        'dest': 'interactive',
    }

    parser = ArgumentParser(prog='dfmpy')
    parser.add_argument(*verbose_args, **verbose_kwargs, dest='main_verbosity')
    subparsers = parser.add_subparsers(help='sub-command help')

    init_parser = subparsers.add_parser('init',
                                        aliases=tuple(['i']),
                                        help='initialize help')
    init_parser.set_defaults(func=initialize)
    init_parser.add_argument(*verbose_args,
                             **verbose_kwargs,
                             dest='verbosity')
    init_parser.add_argument(*force_args,
                             **force_kwargs,
                             help='Force overwriting of files.')
    init_parser.add_argument(*interactive_args,
                             **interactive_kwargs,
                             help='Interactively overwrite files if there are'
                                  ' conflicts.')

    ls_parser = subparsers.add_parser('list',
                                      aliases=tuple(['ls']),
                                      help='list help')
    ls_parser.set_defaults(func=ls)
    ls_parser.add_argument(*verbose_args, **verbose_kwargs, dest='verbosity')

    sync_parser = subparsers.add_parser('sync',
                                        aliases=tuple(['s']),
                                        help='sync help')
    sync_parser.set_defaults(func=sync)
    sync_parser.add_argument(*verbose_args,
                             **verbose_kwargs,
                             dest='verbosity')
    sync_parser.add_argument(*force_args,
                             **force_kwargs,
                             help='Overwrite files if there are conflicts.')
    sync_parser.add_argument(*interactive_args,
                             **interactive_kwargs,
                             help='Interactively overwrite files if there are'
                                  ' conflicts.')

    remove_parser = subparsers.add_parser('remove',
                                          aliases=tuple(['rm']),
                                          help='remove help')
    remove_parser.set_defaults(func=remove)
    remove_parser.add_argument(*verbose_args,
                               **verbose_kwargs,
                               dest='verbosity')
    remove_parser.add_argument(*force_args,
                               **force_kwargs,
                               help='Force removal of files.')
    remove_parser.add_argument(*interactive_args,
                               **interactive_kwargs,
                               help='Interactively remove of files.')

    args = parser.parse_args()
    if not hasattr(args, 'func'):
        parser.print_help()
        parser.exit()

    # Since a argument cannot exist on both the main parser as well as a
    # sub-parser we need to normalize here.  They technically can both exist,
    # but the sub-parser overwrites any value calculated from parsing the main
    # arguments.
    if 'main_verbosity' in args:
        args.verbosity += args.main_verbosity - 1

    return args


def main():
    # TODO docs
    # TODO unit test?
    args = __parse_arguments()
    __setup_logger(args.verbosity)
    try:
        LOG.debug('Arguments: %s', args)

        if 'interactive' in args \
                and args.interactive \
                and not check_for_interactivity():
            raise RuntimeError(
                'Cannot run interactively while not attached to a TTY.')

        args.func(args)

    except KeyboardInterrupt:
        print('')
        LOG.warning('Leaving dfmpy prematurely.')

    # except Exception as e:
    #     LOG.critical(e)
    #     LOG.exception(e)

    except BaseException as e:
        # LOG.critical(e)
        LOG.exception(e)


if __name__ == '__main__':
    main()



# TODO docs

from collections import OrderedDict
from functools import lru_cache
from logging import getLogger
from os import R_OK
from os import access
from pathlib import Path

from dfmpy.utils.config import get_config
from dfmpy.utils.config import get_ignore_globs

# TODO docs
LOG = getLogger()

# TODO docs
# The cache should never get larger than 2, one for $HOME and one for the repo
# dir.  There should be no need for a cache of any other directory.
__FILE_TREE_CACHE_SIZE = 2


# TODO use more generators, do not create lists/tuples/dicts! Is that plausible?


def __ignore(path):
    # TODO docs
    # TODO unit test
    for pattern in get_ignore_globs():
        if path.match(pattern):
            return True
    return False


def __no_permissions_to_stat(path):
    # TODO docs - the path must exist, but is not accessible.
    # TODO unit test
    # TODO is it possible to remove access() in favor or Path() methods?
    return path.exists() and not access(path, R_OK)


def __find_all_paths(dir_path):
    # TODO docs - returns all files and (directories with markers)
    # TODO unit test
    marker = get_config().marker
    if not dir_path.exists():
        return tuple()
    else:
        for path in dir_path.iterdir():
            if __ignore(path):
                LOG.debug('Ignoring: %s', path)

            elif __no_permissions_to_stat(path):
                LOG.error('No permissions to read: %s', path)

            elif path_is_directory(path):
                if marker in path.name:
                    LOG.debug('Found special directory: %s', path)
                    yield path
                else:
                    LOG.debug('Traversing down directory: %s', path)
                    yield from get_all_paths(path)

            elif path_is_file(path):
                LOG.debug('Found file: %s', path)
                yield path

            elif path_is_symlink(path):
                LOG.debug('Found symlink: %s', path)
                yield path

            else:
                LOG.error('Found unknown node type: %s', path)


def __suffixes():
    # TODO docs
    # TODO unit test
    # Order of precedence:
    #   file.txt##hostname
    #   file.txt##hostname.system
    #   file.txt##systemname
    #   file.txt
    marker = get_config().marker
    delimiter = get_config().delimiter
    hostname = get_config().hostname
    system_type = get_config().system
    return tuple([
        marker + hostname,
        marker + delimiter.join([hostname, system_type]),
        marker + system_type,
    ])


def __normalize_for_phantom_files(file_paths):
    # TODO docs - explain this shit!  explain why it's needed when determining expected files when there only exists a markered file (and not the equivalent non-marked file).
    # TODO use this with EXTREME caution, since it's a hack it causes too many side-effects in other places -- causing misleading bugs.
    # TODO unit test
    # TODO optimize this? so it does not iterate over the list many times! (eg use of "in" keyword).
    file_paths = list(file_paths)
    marker = get_config().marker
    for fp in file_paths:
        file_name = str(fp)
        if marker in file_name:
            marker_index = file_name.find(marker)
            phantom_file_name = file_name[:marker_index]
            phantom_file_path = Path(phantom_file_name)
            if phantom_file_path not in file_paths:
                file_paths.append(phantom_file_path)
    file_paths.sort()
    return tuple(file_paths)


def __normalize_repo_path(file_path):
    # TODO docs
    # TODO unit test
    file_name = str(file_path)
    for suffix in __suffixes():
        file_with_suffix = file_name + suffix
        if Path(file_with_suffix).exists():
            return Path(file_with_suffix)
    return Path(file_path)


@lru_cache(maxsize=__FILE_TREE_CACHE_SIZE)
def get_all_paths(dir_name):
    # TODO docs - returns the same as __find_all_paths()
    # TODO unit test
    paths = []
    paths.extend(__find_all_paths(Path(dir_name)))
    paths.sort()
    return tuple(paths)


@lru_cache(maxsize=__FILE_TREE_CACHE_SIZE)
def get_installed_paths(install_dir, repo_dir):
    # TODO docs
    # TODO unit test
    LOG.debug('Searching for installed paths under: %s', install_dir)
    installed_paths = get_all_paths(install_dir)
    installed_paths = [p for p in installed_paths if path_is_symlink(p)]
    installed_paths = [p for p in installed_paths
                       if str(p.resolve()).startswith(repo_dir)]
    mapping = OrderedDict()
    mapping.update([(p, Path(p.resolve())) for p in installed_paths])
    return mapping


@lru_cache(maxsize=__FILE_TREE_CACHE_SIZE)
def get_expected_paths(install_dir, repo_dir):
    # TODO docs
    # TODO unit test
    LOG.debug('Searching for expected paths under: %s', repo_dir)
    marker = get_config().marker
    repo_paths = get_all_paths(repo_dir)
    repo_paths = __normalize_for_phantom_files(repo_paths)
    repo_paths = [p for p in repo_paths if not __ignore(p)]
    repo_paths = [p for p in repo_paths if marker not in p.name]
    mapping = OrderedDict()
    for rp in repo_paths:
        installed_file = str(rp).replace(repo_dir, install_dir)
        installed_path = Path(installed_file)
        nrp = __normalize_repo_path(rp)
        if nrp.resolve().exists():
            mapping[installed_path] = nrp
    return mapping


def path_is_directory(path):
    # TODO docs - explain why path.is_dir() alone is not safe.
    # TODO unit test
    return path_exists(path) and path.is_dir() and not path.is_symlink()


def path_is_file(path):
    # TODO docs - explain why path.is_file() alone is not safe.
    # TODO unit test
    return path_exists(path) and path.is_file()


def path_is_symlink(path):
    # TODO docs - explain why path.is_symlink() alone is not safe.
    # TODO unit test
    return path_exists(path) and path.is_symlink()


def path_exists(path):
    # TODO docs - explain why path.exists() alone is not safe.
    # TODO unit test
    try:
        # Stat the path, but do not follow symlinks.
        path.lstat()
        return True
    except FileNotFoundError:
        return False



# /usr/bin/env python3

from functools import lru_cache
from pathlib import Path
from shutil import rmtree

from setuptools import Command
from setuptools import find_packages
from setuptools import setup

PROJECT_NAME = 'dfmpy'
PROJECT_VERSION = '0.0.0'


def get_repo_file(*args):
    """
    Joins all the arguments into a single string that will be normalized into an
    absolute file path.  Essentially the path will point to a file within this
    repository.

    :param args: Names to join that make up a relative file path.

    :rtype: Path
    :return: The absolute path to a file within this repository.
    """
    return Path(__file__).resolve().parent.joinpath(*args)


@lru_cache(maxsize=2)
def read_file(filename):
    """
    Reads the specified file and returns the full contents as a single string.

    :param filename: The file to read.

    :rtype: str
    :return: The full contents of the specified file.
    """
    with open(filename) as f:
        return f.read()


def get_dfmpy_packages():
    """
    Finds all packages within this project and only returns the production ready
    ones.  Meaning, test packages will not be included.

    :rtype tuple
    :return: A sequence of package names that will be built into the file
            distribution.
    """
    packages = find_packages()
    packages = [p for p in packages if not p.endswith('_test')]
    return tuple(packages)


def load_requirements(filename):
    """
    Loads the :code:`requirements.tx` dependency file and returns the
    dependencies as a sequence.

    The file can contain comments that follow the pound symbol (eg. '#').
    Comments will be removed, each line is stripped of leading and trailing
    whitespace, and empty lines are deleted.  Other than these basic
    transformations, the requirements are left intact.

    :param str filename: The requirements file to read.

    :rtype: tuple
    :return: A sequence of requirements.
    """
    reqs = read_file(filename).splitlines()
    reqs = [str(x).strip() for x in reqs]
    reqs = [x[:x.find('#')] for x in reqs if '#' in x]
    reqs = [x for x in reqs if len(x)]
    return tuple(reqs)


class CleanCommand(Command):
    """
    A custom clean command that removes any intermediate build directories.
    """

    description = 'Custom clean command that forcefully removes build, dist,' \
                  ' and other similar directories.'
    user_options = []

    def __init__(self, *args, **kwargs):
        """Initialized the custom clean command with a list of directories."""
        super(CleanCommand, self).__init__(*args, **kwargs)
        project_path = Path(__file__).resolve().parent
        self._clean_paths = {'.eggs',
                             'build',
                             PROJECT_NAME + '.egg-info',
                             'dist',
                             }
        self._clean_paths = {project_path.joinpath(p)
                             for p in self._clean_paths}
        self._clean_paths = {d for d in self._clean_paths if d.exists()}

    def initialize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def finalize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def run(self):
        """Performs the actual removal of the intermediate build directories."""
        for d in self._clean_paths:
            print(f'Removing {d}')
            rmtree(d)


setup(name=PROJECT_NAME,
      version=PROJECT_VERSION,
      author='Mike Durso',
      author_email='rbprogrammer@gmail.com',
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: MacOS',
          'Operating System :: POSIX',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Unix',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Version Control :: Git',
          'Topic :: Utilities',
      ],
      cmdclass={'clean': CleanCommand},
      description='Another dotfiles manager.',
      include_package_data=True,
      install_requires=load_requirements('requirements.txt'),
      long_description=read_file(get_repo_file('README.md')),
      long_description_content_type='text/markdown',
      packages=get_dfmpy_packages(),
      package_data={'resources': ['*']},
      scripts=('bin/dfmpy',),
      tests_require=load_requirements('requirements.txt'),
      url='https://gitlab.com/rbprogrammer/dfmpy',
      )
